//! A crate of utilities for storing in-memory timeseries data.
//!
//! **Note**: This is a very experimental crate written mainly for use in
//! [bottom](https://github.com/ClementTsang/bottom), and will currently
//! change a lot. You've been warned.

pub mod data;
pub mod time;
